#pragma once

#include "compiler.h"
#include "types.h"
#include "x86.h"

template <typename T>
class Lock_guard
{
    private:
        T& _lock;

    public:
        ALWAYS_INLINE
        Lock_guard (T &l) : _lock (l)
        {
            _lock.lock();
        }

        ALWAYS_INLINE
        ~Lock_guard()
        {
            _lock.unlock();
        }
};

class Spin_lock
{
    private:
        uint8 val;

    public:
        ALWAYS_INLINE
        Spin_lock() : val (1) {}

        ALWAYS_INLINE
        void lock()
        {
            asm volatile (
                "1: lock; decb %0   ;"
                "   jns 3f          ;"
                "2: pause           ;"
                "   cmpb $0, %0     ;"
                "   jle 2b          ;"
                "   jmp 1b          ;"
                "3:                 ;"
                : "+m" (val) : : "cc");
        }

        ALWAYS_INLINE
        void unlock()
        {
            asm volatile ("movb $1, %0;" : "=m" (val) : :);
        }
};

class Ticket_lock
{
    private:
        uint32 cur,next;

    public:
        ALWAYS_INLINE
        Ticket_lock() : cur (0), next (0) {}

        ALWAYS_INLINE
        void lock()
        {
            uint32 dummy;
            asm volatile (
                "   lock; xadd %3, %0   ;"
                "1: pause               ;"
                "   cmpl %2, %1         ;"
                "   jne 1b              ;"
                : "+m" (cur),
                  "=r" (dummy)
                :  "m" (next),
                   "1" (1)
                : "memory", "cc");
        }

        ALWAYS_INLINE
        void unlock()
        {
            asm volatile ("incl %0;" : "+m" (next) : : "memory", "cc");
        }
};

template <typename T>
ALWAYS_INLINE
static T xchg (T& p, T v)
{
    asm volatile ("xchg %1, %0" : "+m" (p), "+r" (v) : :);
    return v;
}

template <typename T>
ALWAYS_INLINE
static T cmpxchg (T& p, T o, T n)
{
    asm volatile ("lock; cmpxchg %2, %0" : "+m" (p), "+a" (o) : "r" (n) : "memory", "cc");
    return o;
}

template <typename T>
ALWAYS_INLINE
static void spin (T& lock)
{
    asm volatile (
        "1: pause               ;"
        "   cmpl    $0, %0      ;"
        "   je      1b          ;"
        : : "m" (lock) : "cc" );
}

class MCS_lock;

class MCS_queue
{
    friend class MCS_lock;

    private:
        MCS_lock* head;

    public:

        MCS_queue() : head (0) {}
};

class MCS_lock
{
    private:
        MCS_lock * volatile _next;
        unsigned volatile _lock;
        MCS_queue & _queue;

    public:
        ALWAYS_INLINE
        MCS_lock (MCS_queue& q) : _queue (q) {}

        ALWAYS_INLINE
        void lock()
        {
            _next = 0;                  // no successor, we are the last
            _lock = 0;                  // prepare spinning
            MCS_lock* j = xchg (_queue.head, this);
            if (j) {
                j->_next = this;        // finally enqueue in list
                spin (_lock);           // wait for other to unlock me ( while (!i->_lock) ; with asm "pause" instruction )
            }
        }

        ALWAYS_INLINE
        void unlock()
        {
            if (!_next) {               // no next?, maybe we are alone
                if (cmpxchg (_queue.head, this, MCS_lock::null()) == this) {
                    return;
                }
                spin (_next);
            }
            _next->_lock = 1;           // release other's lock
        }

        ALWAYS_INLINE
        static MCS_lock* null()         // typed null pointer
        {
            return 0;
        }

};
