/*
 * NOVA system-call interface.
 *
 * Copyright (C) 2008, Bernhard Kauer <bk@vmmon.org>
 * Economic rights: Technische Universitaet Dresden (Germany)
 *
 * This file is part of Vancouver.
 *
 * Vancouver.nova is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * Vancouver.nova is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License version 2 for more details.
 */
#pragma once

#include "types.h"
#include "compiler.h"

// --- constants ---

unsigned const CPU_MAX = 32;

enum
{
    NOVA_CALL,
    NOVA_REPLY,
    NOVA_CREATE_PD,
    NOVA_CREATE_EC,
    NOVA_CREATE_SC,
    NOVA_CREATE_PT,
    NOVA_CREATE_SM,
    NOVA_REVOKE,
    NOVA_LOOKUP,
    NOVA_RECALL,
    NOVA_SEMCTL,
    NOVA_ASSIGN_PCI,
    NOVA_ASSIGN_GSI,
    NOVA_FLAG0          = 1 << 8,
    NOVA_FLAG1          = 1 << 9,
    NOVA_CREATE_ECWORK  = NOVA_CREATE_EC| NOVA_FLAG0,
    NOVA_CREATE_EC_LOCAL= NOVA_CREATE_ECWORK,
    NOVA_SEMCTL_UP      = NOVA_SEMCTL,
    NOVA_SEMCTL_DOWN    = NOVA_SEMCTL | NOVA_FLAG0,
    NOVA_REVOKE_SELF    = NOVA_REVOKE | NOVA_FLAG0,
};

enum {
    NOVA_SUCCESS,
    NOVA_TIMEOUT,
    NOVA_BAD_SYS,
    NOVA_BAD_CAP,
    NOVA_BAD_MEM,
    NOVA_BAD_FTR,
    NOVA_BAD_CPU,
    NOVA_BAD_DEV,
};

enum {
    EXC_UD      = 0x6,
    EXC_GP      = 0xd,
    EXC_PF      = 0xe,
    EXC_STARTUP = 0x1e,
    EXC_RECALL  = 0x1f,
};

// --- common structures ---

class Utcb
{
    public:
    struct head {
        unsigned res0;
        unsigned mtr;
        unsigned crd;
        unsigned res1;
        unsigned res2;
        unsigned tls;
    } head;
    unsigned msg[1024 - sizeof(struct head) / sizeof(unsigned)];
};

class Hip
{   
    public:
        uint32  signature;
        uint16  checksum;       // HIP checksum
        uint16  length;         // HIP length
        uint16  cpu_offs;       // Offset of first CPU descriptor
        uint16  cpu_size;
        uint16  mem_offs;       // Offset of first MEM descriptor
        uint16  mem_size;
        uint32  api_flg;        // API feature flags
        uint32  api_ver;        // API version
        uint32  cfg_cap;        // Number of CAPs
        uint32  cfg_exc;        // Number of Exception portals
        uint32  cfg_vmi;        // Number of VM portals
        uint32  cfg_gsi;        // Number of GSIs
        uint32  cfg_page;       // PAGE sizes
        uint32  cfg_utcb;       // UTCB sizes
        uint32  freq_tsc;       // TSC freq in khz
        uint32  freq_bus;       // BUS freq in khz
};

class Hip_cpu
{
    public:
        uint8   flags;
        uint8   thread;
        uint8   core;
        uint8   package;
        uint32  reserved;
};

class Hip_mem
{
    public:
        uint64  addr;
        uint64  size;
        int32  type;
        uint32  aux;
};

class Crd
{   
    private:
        uint32 _v;

    public:
        enum Type
        {   
            MEM = 1,
            IO  = 2,
            OBJ = 3
        };

        enum Rights
        {   
            NUL = 0,
            R   = 1,
            W   = 2,
            RW  = 3,
            X   = 4,
            RX  = 5,
            WX  = 6,
            RWX = 7,
        };

        ALWAYS_INLINE
        explicit Crd (Type t, uint32 base, uint32 order, Rights r=NUL) : _v (base << 12 | order << 7 | r << 2 | t) {}

        ALWAYS_INLINE
        operator uint32() const
        {
            return _v;
        }
};

class Mtd
{
    private:
        uint32 _v;

    public:
        enum State
        {
            ACDB    = 1u << 0,
            BSD     = 1u << 1,
            ESP     = 1u << 2,
            EIP     = 1u << 3,
            EFL     = 1u << 4,
            DSES    = 1u << 5,
            FSGS    = 1u << 6,
            CSSS    = 1u << 7,
            TR      = 1u << 8,
            LDTR    = 1u << 9,
            GDTR    = 1u << 10,
            IDTR    = 1u << 11,
            CR      = 1u << 12,
            DR      = 1u << 13,
            SYS     = 1u << 14,
            QUAL    = 1u << 15,
            CTRL    = 1u << 16,
            INJ     = 1u << 17,
            STA     = 1u << 18,
            TSC     = 1u << 19,
        };

        ALWAYS_INLINE
        explicit Mtd (uint32 state_or_untyped=0, uint32 typed=0) : _v(typed << 23 | state_or_untyped) {}

        ALWAYS_INLINE
        operator uint32() const
        {
            return _v;
        }
};

class Qpd
{
    private:
        uint32 _v;

    public:
        ALWAYS_INLINE
        explicit Qpd (uint32 time=10000, uint8 prio=1) : _v(time << 12 | prio) {}

        ALWAYS_INLINE
        operator uint32() const
        {
            return _v;
        }
};

static inline unsigned char
syscall (unsigned w0, unsigned w1, unsigned w2, unsigned w3, unsigned w4)
{
    asm volatile (
        "push %%ebp;"
        "mov %%ecx, %%ebp;"
        "mov %%esp, %%ecx;"
        "mov $1f, %%edx;"
        "sysenter;"
        "1: ;"
        "pop %%ebp;"
        : "+a" (w0), "+c"(w4)
        : "D" (w1), "S" (w2), "b"(w3)
        : "edx", "memory");
    return char(w0);
}

extern "C" void NORETURN REGPARAM(1) __reply_and_wait (unsigned, Utcb*);

ALWAYS_INLINE
unsigned char create_ec (unsigned cap, Utcb* utcb, void* stack, unsigned cpu=0, unsigned evt=0, bool local=true)
{
    assert (cpu >> 12 == 0);
    assert ((reinterpret_cast<unsigned>(utcb) & 0xfff) == 0);

    unsigned* _stack = reinterpret_cast<unsigned*>(stack);
    *--_stack = reinterpret_cast<unsigned>(utcb);
    *--_stack = reinterpret_cast<unsigned>(__reply_and_wait);

    return syscall (local ? NOVA_CREATE_EC_LOCAL : NOVA_CREATE_EC, cap, reinterpret_cast<unsigned>(utcb) | cpu, reinterpret_cast<unsigned>(_stack), evt);
}

ALWAYS_INLINE
unsigned char create_ec2 (unsigned cap, Utcb* utcb, void* stack, void REGPARAM(1) (*eip)(unsigned, Utcb*) = 0, unsigned cpu=0, unsigned evt=0, bool local=true)
{
    assert (!local); // going to create a EC+SC, no portal thread
    assert (cpu >> 12 == 0);
    assert ((reinterpret_cast<unsigned>(utcb) & 0xfff) == 0);

    unsigned* _stack = reinterpret_cast<unsigned*>(stack);
    *--_stack = reinterpret_cast<unsigned>(utcb);
    *--_stack = reinterpret_cast<unsigned>(__reply_and_wait);
    *--_stack = reinterpret_cast<unsigned>(eip);

    return syscall (local ? NOVA_CREATE_EC_LOCAL : NOVA_CREATE_EC, cap, reinterpret_cast<unsigned>(utcb) | cpu, reinterpret_cast<unsigned>(_stack), evt);
}

ALWAYS_INLINE
unsigned char create_sc (unsigned cap, unsigned cap_ec, Qpd qpd)
{
    return syscall (NOVA_CREATE_SC, cap, cap_ec, qpd, 0);
}

ALWAYS_INLINE
unsigned char create_pt (unsigned cap, unsigned cap_ec, unsigned REGPARAM(1) (*eip)(unsigned,Utcb*), Mtd mtd=Mtd(0))
{
    return syscall (NOVA_CREATE_PT, cap, cap_ec, mtd, reinterpret_cast<unsigned>(eip));
}

ALWAYS_INLINE
unsigned char call (unsigned cap, Mtd mtd)
{
    return syscall (NOVA_CALL, cap, mtd, 0, 0);
}

ALWAYS_INLINE
unsigned char reply (Mtd mtd)
{
    return syscall (NOVA_REPLY, 0, mtd, 0, 0);
}

ALWAYS_INLINE
unsigned char revoke (unsigned cap, bool self = false)
{
    return syscall (self ? NOVA_REVOKE_SELF : NOVA_REVOKE, cap, 0, 0, 0);
}

ALWAYS_INLINE
unsigned char create_sm (unsigned cap, unsigned counter = 0)
{
    return syscall (NOVA_CREATE_SM, cap, counter, 0, 0);
}

ALWAYS_INLINE
unsigned char sem_up (unsigned cap)
{
    return syscall (NOVA_SEMCTL_UP, cap, 0, 0, 0);
}

ALWAYS_INLINE
unsigned char sem_dn (unsigned cap)
{
    return syscall (NOVA_SEMCTL_DOWN, cap, 0, 0, 0);
}

ALWAYS_INLINE
unsigned char recall (unsigned cap_ec)
{
    return syscall (NOVA_RECALL, cap_ec, 0, 0, 0);
}

// --- asm entry function ---

void main_func (Hip*) asm("main_func") NORETURN;
