#include "nova.h"
#include "printf.h"
#include "lock.h"
#include "x86.h"
#include "string.h"
#include "mem.h"

ALWAYS_INLINE
void check (const unsigned return_code)
{
    if (return_code) FAIL;
}

class Map_self
{
    private:
        Utcb* my_utcb;
        Utcb* mapper_utcb;
        unsigned cap_ec, cap_pt;
        unsigned char mapper_stack[32];

        REGPARAM(1)
        static unsigned mapper (unsigned, Utcb*) { return Mtd (0,0); }

    public:
        ALWAYS_INLINE
        explicit Map_self (Utcb*& utcb, unsigned& caps) : my_utcb(utcb--), mapper_utcb(utcb--), cap_ec(caps++), cap_pt(caps++)
        {   
            // create mapper
            check (create_ec (cap_ec, mapper_utcb, mapper_stack + sizeof(mapper_stack)));
            check (create_pt (cap_pt, cap_ec, mapper));
        }

        ALWAYS_INLINE
        void map (Crd crd)
        {   
            my_utcb->msg[0] = 0;                    // hot spot
            my_utcb->msg[1] = crd;                  // sender window
            mapper_utcb->head.crd = crd;            // receiver window = sender window
            check (call (cap_pt, Mtd(0,1)));
        };
};

class Stack {
    public:
        uint32 data[1024];
};

// i/o functions
unsigned short serial_port;

static inline unsigned char inb (const unsigned port)
{
    unsigned char res;
    asm volatile ("inb %1, %0" : "=a"(res): "Nd"(static_cast<unsigned short>(port)));
    return res;
}

static inline void outb (const unsigned port, const unsigned char value)
{
    asm volatile ("outb %0,%1" :: "a"(value),"Nd"(static_cast<unsigned short>(port)));
}

static inline void init_port(unsigned short const base)
{
    outb (base+3, 0x80);    // LCR : set LDAB = 1
    outb (base  , 1);       // set speed 115k: Divisor Latch Low Byte
    outb (base+1, 0);       // set speed 115k: Divisor Latch High Byte
    outb (base+3, 3);       // LCR : set LDAB=0, 8 bit, 1 stop bit, no parity
    outb (base+1, 1);       // IER : Enable Received Data Available Interrupt
    outb (base+2, 7);       // FCR : enable and clear FIFOs
    // TODO 3 or 11, that means "Aux Output 2 can be connected with the external circuit controlling UART-CPU interrupt process."
    outb (base+4, 3);       // MCR : force DTR, RTS; OUT2 enabled
}

void send (unsigned short const base, const unsigned char value)
{
    while (!(inb(base+5) & 0x20)) ; // LSR : Transmitter Holding Register empty?
    outb (base, value);
}

// function needed as printf backend

void char_out (const char c)
{
    if (!serial_port)
        return;

    if (c == '\n')
        send(serial_port, '\r');

    send (serial_port, c);
}

// startup exception handler, sets the eip to ret, the real eip is on the stack
REGPARAM(1)
static unsigned st_handler (unsigned id, Utcb* utcb)
{
    mword eip;
    asm volatile (
        "   jmp 2f;"
        "1: ret;"
        "2: mov $1b, %0;"
        : "=r" (eip)
    );

    utcb->msg[9] = eip;
    return Mtd(Mtd::EIP);
}

extern void (*CTORS_S)();
extern void (*CTORS_E)();

void main_func (Hip* hip)
{
    // static constructors
    for (void (**func)() = &CTORS_E; func != &CTORS_S; (*--func)()) ;

    unsigned cpu_max = (hip->mem_offs - hip->cpu_offs) / hip->cpu_size;
    unsigned cpu_on = 0;

    assert (cpu_max <= CPU_MAX);

    Hip_cpu* cpu = reinterpret_cast<Hip_cpu*>(reinterpret_cast<char*>(hip) + hip->cpu_offs);

    for (unsigned i = 0; i < cpu_max; i++ )
        cpu_on += cpu[i].flags & 1;

    // where is the first usable capability index
    unsigned caps = cpu_on * hip->cfg_exc + 3;

    // UTCB area, our own UTCB and all the other below it
    Utcb* const my_utcb = reinterpret_cast<Utcb*> (reinterpret_cast<unsigned char*>(hip) - 0x1000);
    Utcb* utcb = my_utcb;

    Map_self self (utcb, caps);

    // map all io ports
    self.map (Crd (Crd::IO, 0, 16, Crd::RWX));

    // map 0-4 kB phys
    Crd crd (Crd::MEM, 0, 0, Crd::R);
    self.map (crd);

    serial_port = *reinterpret_cast<unsigned short*>(0x0400);

    // revoke page
    check (revoke (crd, true));

    if (serial_port)
        init_port (serial_port);

    printf ("%u CPUs online\n",cpu_on);

    // find biggest chunk of free physical memory
    Hip_mem* mem = reinterpret_cast<Hip_mem*> (reinterpret_cast<char*>(hip) + hip->mem_offs);
    int num_mem = (hip->length - hip->mem_offs) / hip->mem_size;

    uint32 addr = 0;
    uint32 size = 0;
    uint32 blk  = 0x400000;     // scan for 4MB chuck at least

    assert ((blk & (blk - 1)) == 0);  // power of two

    for (int i = 0; i < num_mem; i++) {
        printf("mem map addr=%#llx size=%#llx type=%d\n", mem[i].addr, mem[i].size, mem[i].type);
        if ((mem[i].type == 1) && (mem[i].size > size)) {
            addr = static_cast<uint32>(mem[i].addr);
            size = static_cast<uint32>(mem[i].size);
        }
    }

    if (size < blk) {
        printf ("no big enought mem region found\n");
        FAIL;
    }

    printf ("use chunk addr=%#x size=%#x\n",addr,size);

    // ajust alignment of addr to that of blk 
    if (addr & (blk - 1)) {
        uint32 max = addr + size;
        addr += (addr + blk - 1) & ~(blk - 1);
        addr &= blk;
        if (addr + blk >= max) {
            printf ("align addr failed\n");
            FAIL;
        }
    }

    // find free blk-sized and aligned phys mem region
    bool scan = 1;
    while (scan) {
        scan = 0;
        for (int i = 0; i < num_mem; i++) {
            if (mem[i].addr >= addr && mem[i].addr < addr + blk && mem[i].type != 1) {
                uint32 delta = (static_cast<uint32>(mem[i].addr + mem[i].size) - addr + blk - 1);
                delta &= ~(blk - 1);
                if (delta < size) {
                    addr += delta;
                    size -= delta;
                    scan = 1;
                }
            }
        }
    }

    printf ("map phys=virt : %#x - %#x (%#x bytes)\n", addr, addr + blk, blk);

    self.map (Crd (Crd::MEM, addr >> 12, 12, Crd::RW));

    Stack* stack = reinterpret_cast<Stack*>(addr + sizeof(Stack));

    printf ("program complete, enter when ready ...\n");

    while (1) ;

/*
    // startup handler
    check (create_ec (caps, utcb--, stack++, 0, 0, true));
    check (create_pt (EXC_STARTUP, caps++, st_handler));
*/

}
